! This program takes an input file with time-data and calculates the
! powerspectrum/fourier transform of the autocorrelation function.
! The input file must only have lines of the form
!
! <integer> <dummy 1> <dummy 2> <dummy 3> <data 1> <data 2> <data 3>
!
! where the first four entries are ignored and the autocorrelation/
! fourier-transform is performed on the last 3 entries of a line.
! The name of the input file as well as other needed quantities
! are queried by the program interactively.
!
! The output file will contain the power-spectrum up to a maximum frequency.
! The format of this file is:
!
! <frequency> <value1> <value2> <value3> <value4> <value5>
! 
! <frequency> is the frequency in wave numbers.
! <value1>    is the plain power spectrum of the input data (normalized to
!             unity in the output frequency range.
! <value2>    is the power spectrum with a prefactor of 
!             \omega ( 1 - \exp(-\beta \hbar \omega) )
!             corresponding to the classical/Gordon limit.
! <value3>    is the power spectrum with a prefactor of
!             \omega \tanh(\beta \hbar \omega/2)
!             corresponding to the Kubo correction
! <value4>    is the power spectrum with a prefactor of
!             \omega \beta \hbar \omega
!             corresponding to the high temperature / harmonic limit
!             Note: this is the recommended correction factor.
! <value5>    is the power spectrum with a prefactor of
!             \omega ( 1 - \exp(-\beta \hbar \omega) ) *
!                                              \exp(\beta \hbar \omega /2)
!             corresponding to Schofield's correction
! All spectra with their corresponding prefactor are separately normalized
! in the output range to sum up to unity.
!
! Note: the index of refraction of the medium is set to unity.
!
! Version 30.4.2002 - H. Forbert   harald.forbert@theochem.ruhr-uni-bochum.de
! Version 21.2.2006 - fixed off by one allocation error reported
!                     by Michelle Shaw
!
PROGRAM fourier
  
  USE read_data, ONLY : check_file, read_file
  USE kinds, ONLY : dbl, filename_type
  USE fourier_sums, ONLY : fourier_sum

  IMPLICIT NONE

  INTEGER :: n,d,nn,i,j,k,ios, nstep, nover
  TYPE ( filename_type ) :: filename
  TYPE ( filename_type ) :: out_filename
  REAL (dbl) :: a,b,c,w
  REAL (dbl), ALLOCATABLE, DIMENSION ( : , : ) :: input_data
  REAL (dbl), ALLOCATABLE, DIMENSION ( : ) :: ftrans
  REAL (dbl), ALLOCATABLE, DIMENSION ( : ) :: wtrans

  REAL (dbl) :: t, f, e, s, pi, dt
  REAL (dbl) :: data_t,data_temp,max_wave
  REAL (dbl) :: wave_fac, bh

  REAL (dbl) :: norm_fourier, norm_classic, norm_kubo, norm_harmonic
  REAL (dbl) :: norm_shofield

  pi = ACOS(-1.0_dbl)

  PRINT *,"Time-data input filename:"
  READ (*,*) filename%name

  PRINT *,"Output filename:"
  READ (*,*) out_filename%name

  PRINT *,"Time difference in input file (in atomic units):"
  PRINT *,"[CAUTION: not necessarily the MD time step.]"
  READ (*,*) data_t
  wave_fac = 219474.0_dbl/data_t

  PRINT *,"Temperature (in Kelvin):"
  READ (*,*) data_temp
  bh = 1.05459e-34_dbl/1.38066e-23_dbl/2.41889e-17_dbl/data_t/data_temp

  PRINT *,"Output Max Frequency (Wavenumbers):"
  READ (*,*) max_wave

  PRINT *,"Integer frequency increment for output"
  PRINT *,"1 = finest grid/direct fourier transform - standard deviation 100%"
  PRINT *,"n = every n-th frequency - standard deviation about 1/sqrt(n)"
  PRINT *,"Larger numbers result in fewer output frequencies, but also a"
  PRINT *,"less noisy signal"
  READ (*,*) nstep

  IF (nstep < 1) THEN
     PRINT *,"NSTEP must be bigger or equal 1."
     STOP
  END IF

  nover = nstep

!  PRINT *,"Oversampling frequency - e.g. how many frequencies should be"
!  PRINT *,"averaged/summed over. If unsure, use the same number as above for"
!  PRINT *,"the increment."
!  READ (*,*) nover
!
!  IF (nover < 1) THEN
!     PRINT *,"NOVER must be bigger or equal 1."
!     STOP
!  END IF

  CALL check_file(filename,n,d)

  ALLOCATE( input_data ( n+1 , 3*d ), STAT = ios)
  IF (ios /= 0) THEN
     PRINT *,"Out of memory, while trying to allocate array 'input_data'."
     STOP
  END IF

  CALL read_file(input_data,filename,n,3*d)

  nn = INT(n*max_wave/wave_fac/(2.0_dbl*pi))
  IF (nn+1 > n) THEN
     PRINT *,"Maximum wave number too big"
     STOP
  END IF
  nn = nn/nstep
  PRINT *,"Number of frequencies: ",nn

  ALLOCATE( ftrans ( nn+1 ), STAT = ios)
  IF (ios /= 0) THEN
     PRINT *,"Out of memory, trying to allocate 'ftrans'."
     STOP
  END IF

  ALLOCATE( wtrans ( nn+1 ), STAT = ios)
  IF (ios /= 0) THEN
     PRINT *,"Out of memory, trying to allocate 'wtrans'."
     STOP
  END IF

! Welch windowing function:
! Welch function zero at indices 0 and n+1

  input_data ( n+1, : ) = 0.0_dbl
  DO i = 1, n
    input_data(i,:) = input_data(i,:) * (1.0_dbl - ((2*i-n-1.0_dbl)/(n+1))**2)
  END DO

  dt = 2.0_dbl*nstep*pi/((n+1)*nover)

  DO i = 0, nn

!    Base frequency

     t= 2.0_dbl*i*nstep*pi/(n+1)

     c = 0.0_dbl

     DO k = 0, nover-1

        f = 0.0_dbl
        DO j = 1, 3*d
          CALL fourier_sum(n,input_data(:,j),t+k*dt,e,s)
          f = f + e**2+s**2
        END DO

!       input data should have zero mean...
        IF (i+k == 0) f = 0.0_dbl

!       apply cubic spline correction for input data
!       see for example
!       "Numerische Mathematik 1" by J. Stoer, 5th edition, Chapter 2.3.4
!       page 81.

        s=0.5_dbl*(t+k*dt)
        IF (s > 0.1) THEN
          e=(SIN(s)/s)**4
        ELSE
          e=(1.0_dbl-s**2/6.0_dbl+s**4/120.0_dbl)**4
        END IF
        e = e*3.0_dbl/(1.0_dbl+2.0_dbl*COS(s)**2) 

        c = c + e**2*f

     END DO

!    store result

     wtrans(1+i) = t+0.5_dbl*dt*(nover-1)
     ftrans(1+i) = c

  END DO

  norm_fourier  = 0.0_dbl
  norm_classic  = 0.0_dbl
  norm_kubo     = 0.0_dbl
  norm_harmonic = 0.0_dbl 
  norm_shofield = 0.0_dbl
  DO i = 0, nn
     t = wtrans(1+i)
     f = ftrans(1+i)
     e = t*(1.0_dbl - exp(-bh*t))
     norm_fourier  = norm_fourier  + f
     norm_classic  = norm_classic  + f*e
     norm_kubo     = norm_kubo     + f*e/(1.0_dbl+exp(-bh*t))
     norm_harmonic = norm_harmonic + f*t**2
     norm_shofield = norm_shofield + f*e*exp(0.5_dbl*bh*t)
  END DO 
  c = 1.0_dbl
  norm_fourier  = 1.0_dbl/(norm_fourier *c)
  norm_classic  = 1.0_dbl/(norm_classic *c)
  norm_kubo     = 1.0_dbl/(norm_kubo    *c)
  norm_harmonic = 1.0_dbl/(norm_harmonic*c)
  norm_shofield = 1.0_dbl/(norm_shofield*c)

  OPEN (UNIT = 10, FILE = out_filename%name, STATUS = "UNKNOWN", &
        ACTION = "WRITE", IOSTAT = ios)
  IF (ios /= 0) THEN
     PRINT *,"Failed to open output file!"
     STOP
  END IF

  DO i = 0, nn
     t = wtrans(1+i)
     f = ftrans(1+i)
     e = t*(1.0_dbl - exp(-bh*t))
     WRITE (10,'(1X,6ES13.5)',IOSTAT=ios) wave_fac*t,norm_fourier*f, &
        norm_classic *f*e, norm_kubo*f*e/(1.0_dbl+exp(-bh*t)), &
        norm_harmonic*f*t**2, &
        norm_shofield*f*e*exp(0.5_dbl*bh*t)
  END DO 

  CLOSE(10, IOSTAT=ios)

  PRINT *,"Done."

  STOP

END PROGRAM fourier
